from collections.abc import Generator


def splitlines(s: str) -> Generator[str]:
    for line in s.strip().splitlines():
        line_stripped: str = line.strip()
        if line_stripped:
            yield line_stripped



import asyncio
import asyncio.subprocess as asp
import os
import subprocess as sp
from typing import Literal, overload

type StrPath = str | os.PathLike[str]


@overload
async def run(
    *args: StrPath, capture_stdout: Literal[False] = False, text: bool = False
) -> None: ...
@overload
async def run(
    *args: StrPath, capture_stdout: Literal[True], text: Literal[False] = False
) -> bytes: ...
@overload
async def run(
    *args: StrPath, capture_stdout: Literal[True], text: Literal[True]
) -> str: ...
async def run(
    *args: StrPath, capture_stdout: bool = False, text: bool = False
) -> str | bytes | None:
    proc: asp.Process = await asyncio.create_subprocess_exec(
        *[str(a) for a in args], stdout=asp.PIPE if capture_stdout else None
    )
    output: bytes = b""
    if capture_stdout:
        assert proc.stdout is not None
        output = await proc.stdout.read()
    returncode: int = await proc.wait()
    if returncode != 0:
        raise sp.CalledProcessError(returncode, args)
    if capture_stdout:
        if text:
            return output.decode()
        return output
    return None



import os
from pathlib import Path

from . import DEFAULT_HASHER

FILENAMES: dict[str, str] = {
    "blake2b": "b2sums.txt",
}


def sums(hasher: str = DEFAULT_HASHER) -> str:
    if hasher in FILENAMES:
        return FILENAMES[hasher]
    return hasher + "sums.txt"


def single(fpath: str | os.PathLike[str], hasher: str = DEFAULT_HASHER) -> str:
    fpath: Path = Path(fpath)
    return fpath.name + "." + hasher



import logging

from environs import env

# make pyright happy
import liblaf.grapes as grapes  # noqa: PLR0402


def init_logging(level: str | int = logging.NOTSET) -> None:
    if level in (logging.NOTSET, "NOTSET"):
        level = "DEBUG" if env.bool("RUNNER_DEBUG", False) else "INFO"
    grapes.init_logging(level)



from pathlib import Path
from typing import Any

from environs import Env, env

from liblaf import actions

input_env = Env(prefix="INPUT_")


def get_input(name: str) -> str:
    val: str = input_env.str(name.replace(" ", "_").upper(), "")
    return val.strip()


def get_multiline_input(name: str) -> list[str]:
    return list(actions.utils.splitlines(get_input(name)))


def notice(message: str) -> None:
    print(f"::notice::{message}")


def set_output(name: str, value: Any, *, delimiter: str = "EOF") -> None:
    fpath: Path = env.path("GITHUB_OUTPUT")
    value = str(value)
    with fpath.open("a") as fp:
        if "\n" in value:
            # ref: <https://docs.github.com/en/actions/writing-workflows/choosing-what-your-workflow-does/workflow-commands-for-github-actions#multiline-strings>
            fp.write(f"{name}<<{delimiter}\n{value}\n{delimiter}\n")
        else:
            fp.write(f"{name}={value}\n")



import functools
import hashlib
import os
from pathlib import Path

from . import DEFAULT_HASHER


def hash_bytes(data: bytes, hasher: str = DEFAULT_HASHER) -> str:
    hasher: hashlib._Hash = hashlib.new(hasher)
    hasher.update(data)
    return hasher.hexdigest()


@functools.lru_cache
def hash_file(fpath: str | os.PathLike[str], hasher: str = DEFAULT_HASHER) -> str:
    fpath = Path(fpath)
    with fpath.open("rb") as fp:
        hasher: hashlib._Hash = hashlib.file_digest(fp, hasher)
        return hasher.hexdigest()


def hash_files(
    *files: str | os.PathLike[str], hasher: str = DEFAULT_HASHER
) -> dict[str, str]:
    result: dict[str, str] = {}
    for _file in files:
        fpath = Path(_file)
        result[fpath.name] = hash_file(fpath, hasher)
    return result



import functools

import pydantic_settings as ps
from loguru import logger

from liblaf.actions import core


class Inputs(ps.BaseSettings):
    model_config = ps.SettingsConfigDict(env_prefix="INPUT_")

    repo: str

    @functools.cached_property
    def author(self) -> list[str]:
        authors: list[str] = core.get_multiline_input("AUTHOR")
        logger.info("Authors:\n{}", "\n".join(authors))
        return authors



import githubkit
from environs import env


def get_octokit(*, http_cache: bool = True) -> githubkit.GitHub:
    token: str | None = (
        env.str("INPUT_TOKEN", None)
        or env.str("INPUT_GITHUB_TOKEN", None)
        or env.str("GH_TOKEN", None)
        or env.str("GITHUB_TOKEN", None)
    )
    return githubkit.GitHub(token, http_cache=http_cache)



from . import cksum
from ._action import action
from ._logging import init_logging
from ._subprocess import run
from ._text import splitlines

__all__ = [
    "action",
    "cksum",
    "init_logging",
    "run",
    "splitlines",
]



from liblaf.actions import core, utils

from . import Client, Inputs


@utils.action()
async def main(inputs: Inputs) -> None:
    client = Client(*inputs.repo.split("/"))
    create: bool = False
    cksums_local: dict[str, str] = utils.cksum.hash_files(
        *inputs.files, hasher=inputs.hasher
    )
    if await client.release_exists(inputs.tag):
        cksums_remote: dict[str, str] = await client.release_cksums(
            inputs.tag, inputs.hasher
        )
        if cksums_local == cksums_remote:
            core.notice(f"Hashsums match, skip release: {inputs.tag}")
            return
        if inputs.clobber:
            core.notice(f"Recreate release: {inputs.tag}")
            await client.release_delete(inputs.tag)
            create = True
        else:
            core.notice(f"Update release: {inputs.tag}")
            await client.release_upload(inputs.tag, *inputs.files, hasher=inputs.hasher)
            await client.release_update(inputs.tag, body=inputs.changelog)
    else:
        create = True
        core.notice(f"Create release: {inputs.tag}")
    if create:
        await client.release_create(
            inputs.tag,
            *inputs.files,
            hasher=inputs.hasher,
            notes=inputs.changelog,
            prerelease=inputs.prerelease,
        )



import asyncio
import functools
import inspect
from collections.abc import Callable
from typing import Generic, Protocol, TypeVar

from pydantic_settings import BaseSettings

from liblaf.actions import utils

_S_contra = TypeVar("_S_contra", bound=BaseSettings, contravariant=True)


class RawAction(Generic[_S_contra], Protocol):
    async def __call__(self, inputs: _S_contra) -> None: ...


class WrappedAction(Protocol):
    def __call__(self) -> None: ...


def action() -> Callable[[RawAction], WrappedAction]:
    def decorator(fn: RawAction[_S_contra]) -> WrappedAction:
        @functools.wraps(fn)
        def wrapped() -> None:
            utils.init_logging()
            sig: inspect.Signature = inspect.signature(fn)
            param: inspect.Parameter = next(iter(sig.parameters.values()))
            annotation: type[_S_contra] = param.annotation
            inputs: _S_contra = annotation()
            asyncio.run(fn(inputs))

        return wrapped

    return decorator



import itertools
from collections.abc import AsyncGenerator
from typing import Any

import githubkit
import githubkit.versions.latest.models as m


class AppClient:
    _gh: githubkit.GitHub

    def __init__(self, gh: githubkit.GitHub) -> None:
        self._gh = gh

    async def list_installations(self) -> AsyncGenerator[m.Installation, Any]:
        for page in itertools.count(1):
            installations: list[m.Installation] = (
                await self._gh.rest.apps.async_list_installations(page=page)
            ).parsed_data
            if not installations:
                break
            for installation in installations:
                yield installation

    async def list_repos_accessible_to_installation(
        self,
    ) -> AsyncGenerator[m.Repository, Any]:
        total_count: int = 0
        for page in itertools.count(1):
            resp: m.InstallationRepositoriesGetResponse200 = (
                await self._gh.rest.apps.async_list_repos_accessible_to_installation(
                    page=page
                )
            ).parsed_data
            for repo in resp.repositories:
                yield repo
            total_count += len(resp.repositories)
            if total_count >= resp.total_count:
                break



import os
from pathlib import Path

from liblaf.actions import utils


def parse(text: str | bytes) -> dict[str, str]:
    result: dict[str, str] = {}
    for line in utils.splitlines(text if isinstance(text, str) else text.decode()):
        hashsum: str
        filename: str
        hashsum, filename = line.split()
        result[filename] = hashsum
    return result


def dumps(data: dict[str, str]) -> str:
    text: str = ""
    for filename, hashsum in sorted(data.items()):
        text += f"{hashsum}  {filename}\n"
    return text


def dump(data: dict[str, str], fpath: str | os.PathLike[str]) -> None:
    fpath = Path(fpath)
    fpath.write_text(dumps(data))



import pydantic_settings as ps


class Inputs(ps.BaseSettings):
    model_config = ps.SettingsConfigDict(env_prefix="INPUT_", cli_parse_args=True)

    repo: str
    source_repo: str
    source_ruleset_id: int



from . import filename
from ._constants import DEFAULT_HASHER
from ._hashsum import hash_bytes, hash_file, hash_files
from ._sumfile import dump, dumps, parse

__all__ = [
    "DEFAULT_HASHER",
    "dump",
    "dumps",
    "filename",
    "hash_bytes",
    "hash_file",
    "hash_files",
    "parse",
]



import asyncio
import contextlib
import os
from collections.abc import Generator
from pathlib import Path
from typing import TYPE_CHECKING, Any

import githubkit
import githubkit.exception
import githubkit.typing
import githubkit.utils
import githubkit.versions.latest.models as ghm
import httpx
import tenacity

from liblaf.actions import github
from liblaf.actions.utils import cksum

if TYPE_CHECKING:
    from collections.abc import Coroutine


class Client:
    _gh: githubkit.GitHub
    owner: str
    repo: str

    def __init__(self, owner: str, repo: str) -> None:
        self._gh = github.get_octokit(http_cache=False)
        self.owner = owner
        self.repo = repo

    async def release_get(self, tag: str) -> ghm.Release:
        resp: githubkit.Response[
            ghm.Release
        ] = await self._gh.rest.repos.async_get_release_by_tag(
            self.owner, self.repo, tag
        )
        return resp.parsed_data

    async def release_exists(self, tag: str) -> bool:
        try:
            await self.release_get(tag)
        except githubkit.exception.RequestFailed as err:
            if err.response.status_code == httpx.codes.NOT_FOUND:
                return False
            raise
        else:
            return True

    async def release_create(
        self,
        tag: str,
        *files: str | os.PathLike[str],
        hasher: str | None = None,
        notes: str | None = None,
        prerelease: bool = False,
    ) -> ghm.Release:
        kwargs: dict[str, Any] = (
            {"body": notes, "generate_release_notes": False}
            if notes
            else {"generate_release_notes": True}
        )
        resp: githubkit.Response[
            ghm.Release
        ] = await self._gh.rest.repos.async_create_release(
            self.owner,
            self.repo,
            tag_name=tag,
            name=tag,
            prerelease=prerelease,
            **kwargs,
        )
        release: ghm.Release = resp.parsed_data
        await self._wait_until_release(tag, exists=True)
        await self.release_upload(tag, *files, hasher=hasher)
        return release

    async def release_delete(self, tag: str) -> None:
        with suppress_not_found():
            release: ghm.Release = await self.release_get(tag)
            await self._gh.rest.repos.async_delete_release(
                self.owner, self.repo, release.id
            )
        with suppress_not_found():
            await self._gh.rest.git.async_delete_ref(
                self.owner, self.repo, f"tags/{tag}"
            )
        # TODO: remove workaround for [cli/cli#5024 (comment)](https://github.com/cli/cli/issues/5024#issuecomment-1028018586)
        await self._wait_until_release(tag, exists=False)
        await asyncio.sleep(5)

    async def release_download(self, tag: str, asset_name: str) -> bytes:
        resp: httpx.Response = await self._gh._arequest(  # noqa: SLF001
            "GET",
            f"https://github.com/{self.owner}/{self.repo}/releases/download/{tag}/{asset_name}",
        )
        resp = resp.raise_for_status()
        return resp.content

    async def release_delete_asset(self, tag: str, asset_name: str) -> None:
        with suppress_not_found():
            release: ghm.Release = await self.release_get(tag)
            for asset in release.assets:
                if asset.name == asset_name:
                    await self._gh.rest.repos.async_delete_release_asset(
                        self.owner, self.repo, asset.id
                    )
                    return

    async def release_cksums(self, tag: str, hasher: str) -> dict[str, str]:
        try:
            data: bytes = await self.release_download(tag, cksum.filename.sums(hasher))
        except httpx.HTTPStatusError as err:
            if err.response.status_code == httpx.codes.NOT_FOUND:
                return {}
            raise
        else:
            return cksum.parse(data)

    async def release_upload(
        self, tag: str, *files: str | os.PathLike[str], hasher: str | None = None
    ) -> list[ghm.ReleaseAsset]:
        cksums: dict[str, str]
        if hasher:
            cksums = await self.release_cksums(tag, hasher)
        else:
            cksums = {}
        futures: list[Coroutine[Any, Any, ghm.ReleaseAsset]] = []
        for file in files:
            fpath: Path = Path(file)
            data: bytes = fpath.read_bytes()
            futures.append(self._release_upload_asset(tag, fpath.name, data=data))
            if hasher:
                s: str = cksum.hash_bytes(data, hasher)
                cksums[fpath.name] = s
                # futures.append(
                #     self._release_upload_asset(
                #         tag,
                #         cksum.filename.single(fpath, hasher),
                #         data=cksum.dumps({fpath.name: s}).encode(),
                #     )
                # )
        if hasher:
            futures.append(
                self._release_upload_asset(
                    tag,
                    cksum.filename.sums(hasher),
                    data=cksum.dumps(cksums).encode(),
                )
            )
        return await asyncio.gather(*futures)

    @tenacity.retry(
        wait=tenacity.wait_random_exponential(max=60),
        retry=tenacity.retry_if_exception_type(tenacity.TryAgain),
    )
    async def _wait_until_release(self, tag: str, *, exists: bool) -> None:
        if await self.release_exists(tag) != exists:
            raise tenacity.TryAgain

    async def _release_upload_asset(
        self, tag: str, name: str, *, data: bytes
    ) -> ghm.ReleaseAsset:
        release: ghm.Release = await self.release_get(tag)
        await self.release_delete_asset(tag, name)
        # TODO: replace with `async_upload_release_asset()` until [yanyongyu/githubkit#150](https://github.com/yanyongyu/githubkit/issues/150) is fixed
        resp: githubkit.Response[ghm.ReleaseAsset, Any] = await self._gh.arequest(
            "POST",
            f"https://uploads.github.com/repos/{self.owner}/{self.repo}/releases/{release.id}/assets",
            params={"name": name},
            content=data,
            headers={"Content-Type": "application/octet-stream"},
            response_model=ghm.ReleaseAsset,
        )
        return ghm.ReleaseAsset.model_validate(resp.json())

    async def release_update(
        self, tag: str, *, body: githubkit.typing.Missing[str] = githubkit.utils.UNSET
    ) -> None:
        release: ghm.Release = await self.release_get(tag)
        if not body:
            return
        await self._gh.rest.repos.async_update_release(
            self.owner, self.repo, release.id, body=body
        )


@contextlib.contextmanager
def suppress_not_found() -> Generator[None]:
    try:
        yield
    except githubkit.exception.RequestFailed as err:
        if err.response.status_code != httpx.codes.NOT_FOUND:
            raise



from typing import Any

import githubkit

from liblaf.actions import core, github, utils

from . import Inputs


@utils.action()
async def main(inputs: Inputs) -> None:
    _: Any
    gh: githubkit.GitHub = github.get_octokit()
    owner: str
    repo: str
    owner, _, repo = inputs.repo.partition("/")
    async for pr in gh.paginate(
        gh.rest.pulls.async_list, owner=owner, repo=repo, state="open"
    ):
        if (pr.user is None) or (pr.user.login not in inputs.author):
            continue
        async for review in gh.paginate(
            gh.rest.pulls.async_list_reviews,
            owner=owner,
            repo=repo,
            pull_number=pr.number,
        ):
            if review.state == "APPROVED" and review.commit_id == pr.head.sha:
                break
        else:
            core.notice(f"Approve PR: <{pr.html_url}>")
            await gh.rest.pulls.async_create_review(
                owner=owner, repo=repo, pull_number=pr.number, event="APPROVE"
            )



import functools
import re
from pathlib import Path

import pydantic_settings as ps
from loguru import logger

from liblaf.actions import core


class Inputs(ps.BaseSettings):
    model_config = ps.SettingsConfigDict(env_prefix="INPUT_")

    clobber: bool = False
    hasher: str = "sha256"
    prerelease: bool = False
    repo: str
    tag: str

    @functools.cached_property
    def changelog(self) -> str | None:
        fpath: Path = Path(core.get_input("CHANGELOG_FILE"))
        if not fpath.exists():
            return None
        text: str = fpath.read_text()
        text = re.sub(r"## \[unreleased\]", "", text, flags=re.IGNORECASE)
        text = text.strip()
        if not text:
            return None
        return text

    @functools.cached_property
    def files(self) -> list[Path]:
        files: list[Path] = []
        for line in core.get_multiline_input("FILES"):
            files.extend(Path.cwd().glob(line))
        logger.info("Files:\n{}", "\n".join([str(f) for f in files]))
        return files



import asyncio
from typing import Protocol, get_type_hints

from liblaf import grapes

from . import Inputs


class ActionFunction[T: Inputs](Protocol):
    async def __call__(self, inputs: T) -> None: ...


def run[T: Inputs](func: ActionFunction[T]) -> None:
    grapes.init_logging()
    type_hints: dict[str, type[T]] = get_type_hints(func)
    cls: type[T] = type_hints["inputs"]
    inputs: T = cls()
    asyncio.run(func(inputs))



from typing import Any

import githubkit
import githubkit.versions.latest.models as ghm
import githubkit.versions.latest.types as ght
import pydantic

from liblaf.actions import core, github, utils

from . import Inputs


@utils.action()
async def main(inputs: Inputs) -> None:
    _: Any
    gh: githubkit.GitHub = github.get_octokit()
    owner: str
    repo: str
    owner, _, repo = inputs.repo.partition("/")
    source_owner: str
    source_repo: str
    source_owner, _, source_repo = inputs.source_repo.partition("/")
    source_ruleset: ghm.RepositoryRuleset = (
        await gh.rest.repos.async_get_repo_ruleset(
            source_owner, source_repo, inputs.source_ruleset_id
        )
    ).parsed_data
    target_ruleset: ghm.RepositoryRuleset | None = await find_ruleset(
        owner, repo, source_ruleset.name
    )
    if target_ruleset is not None:
        await update_ruleset(owner, repo, source_ruleset, target_ruleset.id)
    else:
        await create_ruleset(owner, repo, source_ruleset)


async def find_ruleset(
    owner: str, repo: str, name: str
) -> ghm.RepositoryRuleset | None:
    gh: githubkit.GitHub = github.get_octokit()
    async for ruleset in gh.paginate(
        gh.rest.repos.async_get_repo_rulesets, owner=owner, repo=repo
    ):
        if ruleset.name == name:
            return (
                await gh.rest.repos.async_get_repo_ruleset(owner, repo, ruleset.id)
            ).parsed_data
    return None


async def create_ruleset(owner: str, repo: str, ruleset: ghm.RepositoryRuleset) -> None:
    gh: githubkit.GitHub = github.get_octokit()
    adapter = pydantic.TypeAdapter(ght.ReposOwnerRepoRulesetsPostBodyType)
    data: ght.ReposOwnerRepoRulesetsPostBodyType = adapter.validate_python(
        ruleset.model_dump(exclude_unset=True, exclude_defaults=True, exclude_none=True)
    )
    core.notice(f'Create Ruleset "{ruleset.name}".')
    await gh.rest.repos.async_create_repo_ruleset(owner, repo, data=data)


async def update_ruleset(
    owner: str, repo: str, ruleset: ghm.RepositoryRuleset, target_ruleset_id: int
) -> None:
    gh: githubkit.GitHub = github.get_octokit()
    adapter = pydantic.TypeAdapter(ght.ReposOwnerRepoRulesetsRulesetIdPutBodyType)
    data: ght.ReposOwnerRepoRulesetsRulesetIdPutBodyType = adapter.validate_python(
        ruleset.model_dump(exclude_unset=True, exclude_defaults=True, exclude_none=True)
    )
    core.notice(f'Update Ruleset "{ruleset.name}".')
    await gh.rest.repos.async_update_repo_ruleset(
        owner, repo, target_ruleset_id, data=data
    )
